// ===========================================================================
// Buffer.h
// Memory management
//
//                                                                    kanegon
// ===========================================================================

#if !defined(_BUFFER_H_)
#define _BUFFER_H_

#include <windows.h>
#include <stdexcept>

/////////////////////////////////////////////////////////////////////////////

#define GROW_INIT    32        // initialize size
#define GROW_DELTA   0x100000  // delta size for grow (about 1Mbyte)


/////////////////////////////////////////////////////////////////////////////

// throws exception
// std::invalid_argument
// std::bad_alloc

class CBuffer {
private:
	BYTE* m_pBuffer;
	DWORD m_dwBufferSize;
	DWORD m_dwBufferUsed;
	DWORD m_dwReadCursor;
	BOOL m_bAttached;

public:
	CBuffer() {
		m_pBuffer = NULL;
		m_bAttached = FALSE;
		Clear();
	}
	CBuffer(BYTE* pBuffer, DWORD dwBufferSize) {
		m_pBuffer = NULL;
		Attach(pBuffer, dwBufferSize);
	}
	virtual ~CBuffer() {
		Clear();
	}

private:
	void GrowBuffer(DWORD dwRequireSize = 0) {
		if (m_bAttached) {
			throw std::bad_alloc();
		}
		if (m_pBuffer == NULL) {
			DWORD dwSize = max(dwRequireSize, GROW_INIT);
			m_pBuffer = new BYTE[dwSize];
			if (m_pBuffer == NULL) {
				throw std::bad_alloc();
			}
			m_dwBufferSize = dwSize;
			m_dwBufferUsed = 0;
			m_dwReadCursor = 0;
			return;
		}
		DWORD dwNewSize;
		if (m_dwBufferSize < GROW_DELTA) {
			dwNewSize = m_dwBufferSize * 2;
		} else {
			dwNewSize = m_dwBufferSize + GROW_DELTA;
		}
		if (dwNewSize - m_dwBufferUsed < dwRequireSize) {
			dwNewSize = dwNewSize + dwRequireSize;
		}
		BYTE* pNewBuffer = new BYTE[dwNewSize];
		if (pNewBuffer == NULL) {
			throw std::bad_alloc();
		}
		memcpy(pNewBuffer, m_pBuffer, m_dwBufferSize);
		delete[] m_pBuffer;
		m_pBuffer = pNewBuffer;
		m_dwBufferSize = dwNewSize;
	}

public:
	void Clear() {
		if (m_pBuffer != NULL) {
			if (! m_bAttached) {
				delete[] m_pBuffer;
			}
		}
		m_pBuffer = NULL;
		m_dwBufferSize = 0;
		m_dwBufferUsed = 0;
		m_dwReadCursor = 0;
		m_bAttached = FALSE;
	}
	void Attach(BYTE* pBuffer, DWORD dwBufferSize) {
		Clear();
		if (pBuffer == NULL) {
			return;
		}
		m_pBuffer = pBuffer;
		m_dwBufferSize = dwBufferSize;
		m_dwBufferUsed = dwBufferSize;
		m_bAttached = TRUE;
	}
	void Detach(BYTE** ppBuffer, DWORD* pdwBufferSize) {
		if (m_pBuffer == NULL) {
			*ppBuffer = NULL;
			*pdwBufferSize = 0;
			return;
		}
		*ppBuffer = m_pBuffer;
		*pdwBufferSize = m_dwBufferUsed;
		m_pBuffer = NULL;
		Clear();
	}
	BYTE* GetBuffer() {
		return m_pBuffer;
	}
	DWORD GetBufferSize() {
		return m_dwBufferUsed;
	}
	BOOL IsEof() {
		if (m_pBuffer == NULL) {
			return TRUE;
		}
		if (m_dwReadCursor >= m_dwBufferUsed) {
			return TRUE;
		}
		return FALSE;
	}
	BOOL Read(BYTE* pBuffer, DWORD dwReadSize) {
		if (pBuffer == NULL) {
			throw std::invalid_argument("null pointer exception");
		}
		if (dwReadSize == 0) {
			return TRUE;
		}
		if (IsEof()) {
			return FALSE;
		}
		if (m_dwBufferUsed - m_dwReadCursor < dwReadSize) {
			return FALSE;
		}
		memcpy(pBuffer, &m_pBuffer[m_dwReadCursor], dwReadSize);
		m_dwReadCursor += dwReadSize;
		return TRUE;
	}
	BOOL Read(BYTE* pBuffer, DWORD dwBufferSize, DWORD* pdwReadSize) {
		if (pBuffer == NULL || pdwReadSize == NULL) {
			throw std::invalid_argument("null pointer exception");
		}
		if (dwBufferSize == 0) {
			return TRUE;
		}
		if (IsEof()) {
			return FALSE;
		}
		DWORD dwReadSize = min(dwBufferSize, m_dwBufferUsed - m_dwReadCursor);
		memcpy(pBuffer, &m_pBuffer[m_dwReadCursor], dwReadSize);
		m_dwReadCursor += dwReadSize;
		*pdwReadSize = dwReadSize;
		return TRUE;
	}
	void Write(const BYTE* pBuffer, DWORD dwWriteSize) {
		if (dwWriteSize == 0) {
			return;
		}
		if (pBuffer == NULL) {
			throw std::invalid_argument("null pointer exception");
		}
		if (m_dwBufferSize - m_dwBufferUsed < dwWriteSize) {
			GrowBuffer(dwWriteSize);
		}
		memcpy(&m_pBuffer[m_dwBufferUsed], pBuffer, dwWriteSize);
		m_dwBufferUsed += dwWriteSize;
	}
};

/////////////////////////////////////////////////////////////////////////////

class CTextBuffer : public CBuffer {
public:
	CTextBuffer() {
	}
	CTextBuffer(char* pszText) : CBuffer((BYTE*)pszText, lstrlen(pszText) + 1) {
	}
	~CTextBuffer() {
	}

public:
	void WriteText(const char* pszText) {
		Write((BYTE*)pszText, lstrlen(pszText));
	}
	void WriteLine(const char* pszText = NULL) {
		if (pszText != NULL) {
			Write((BYTE*)pszText, lstrlen(pszText));
		}
		Write((BYTE*)"\r\n", 2);
	}
	void WriteNull(const char* pszText = NULL) {
		if (pszText != NULL) {
			Write((BYTE*)pszText, lstrlen(pszText));
		}
		Write((BYTE*)"", 1);
	}
};

#endif // _BUFFER_H_
